"""
Build an object to represent a Clarisse Conductor submission.

Also handle the submit action.

NOTE There are some bugs in clarisse that have lead to some decisions for the
submission flow.

1. Bug where some nested path overrides don't correctly get overridden when
   opening the project.

2. Undo doesn't work on deeply nested reference contexts. If you make-all-local
   in a batch undo block, then when you undo, there are extra nodes in the
   project.

Users can choose to localize contexts before submission (off by default).
If they turn it on then the flow is:
*. Make sure the user has saved the file. (they might need it)
*. Make contexts local.
*. Remove Conductor nodes.
*. Save the render package.
*. Do the submission.
*. Reload the saved file. This is slow for very big projects because of the time
   to reload the project.

If they don't localize, then:
*. Save the render package.
*. Do the submission.

In this case the submission may contain xrefs nested to any level and we do a
pretty good job of resolving them. However, if the render errors due to xref
the localize method is the fallback.
"""

import datetime
import errno
import fileinput
import os
import re
import shutil
import sys
import tempfile
import traceback

import cioclarisse.clarisse_config as ccfg
import cioclarisse.scripted_class.dependencies as deps
import cioclarisse.utils as cu
import ix
from cioclarisse.scripted_class import missing_files_ui
from cioclarisse.scripted_class.job import Job
from ciocore import conductor_submit
from cioclarisse.utils import ConductorError
from ciocore import data as coredata

from ciocore.gpath import Path
from ciocore.gpath_list import PathList

PROJECT_EXTENSION_REGEX = r"(\.ct\.project|\.project)"
CT_PROJECT_EXTENSION = ".ct.project"

FILE_ATTR_HINTS = [
    ix.api.OfAttr.VISUAL_HINT_FILENAME_SAVE,
    ix.api.OfAttr.VISUAL_HINT_FILENAME_OPEN,
    ix.api.OfAttr.VISUAL_HINT_FOLDER,
]


def _get_path_line_regex():
    """
    Generate a regex to help identify filepath attributes.

    As we scan project files to replace windows paths, we use this regex which
    will be something like: r'\s+(?:filename|filename_sys|save_as)\s+"(.*)"\s+' 
    only longer.
    """
    classes = ix.application.get_factory().get_classes()
    file_attrs = []
    for klass in classes.get_classes():
        attr_count = klass.get_attribute_count()
        for i in xrange(attr_count):
            attr = klass.get_attribute(i)
            hint = attr.get_visual_hint()
            if hint in FILE_ATTR_HINTS:
                file_attrs.append(attr.get_name())

    return r"\s+(?:" + "|".join(sorted(set(file_attrs))) + r')\s+"(.*)"\s+'


def _localize_contexts():
    """
    Make all clarisse reference contexts local.

    This is the equivalent of importing Maya references and is optional for the
    user.
    """
    contexts = ix.api.OfContextSet()
    ix.application.get_factory().get_root().resolve_all_contexts(contexts)
    for ctx in contexts:
        if ctx.is_reference() and ctx.get_attribute("filename").get_string().endswith(
            ".project"
        ):
            ix.cmds.MakeLocalContext(ctx)


def _remove_conductor():
    """
    Remove all Conductor data from the render archive.

    This ensures the render logs are not polluted by complaints about Conductor
    nodes. This can only be done in the situation where we localize contexts,
    because in that case we get to reload the scene after submission.
    """
    objects = ix.api.OfObjectArray()
    ix.application.get_factory().get_objects("ConductorJob", objects)
    for item in list(objects):
        ix.application.get_factory().remove_item(item.get_full_name())


class Submission(object):
    """
    Submission holds all data needed for a submission.

    It has potentially many Jobs, and those Jobs each have many Tasks. A
    Submission can provide the correct args to send to Conductor, or it can be
    used to create a dry run to show the user what will happen.

    A Submission also sets a list of tokens that the user can access as <angle
    bracket> tokens in order to build strings in the UI such as commands, job
    title, and (soon to be added) metadata.
    """

    def __init__(self, obj):
        """
        Collect data from the Clarisse UI.

        Collect attribute values that are common to all jobs, then call
        _set_tokens(). After _set_tokens has been called, the Submission level
        token variables are valid and calls to evaluate expressions will
        correctly resolve where those tokens have been used.
        """
        self.node = obj

        if self.node.is_kindof("ConductorJob"):
            self.nodes = [obj]
        else:
            raise NotImplementedError

        self.localize_before_ship = self.node.get_attribute(
            "localize_contexts"
        ).get_bool()
        self.project_filename = ix.application.get_current_project_filename()
 
        self.tmpdir = Path(
            os.path.join(
                ix.application.get_factory().get_vars().get("CTEMP").get_string(),
                "conductor",
            )
        )
        self.render_package_path = self._get_render_package_path()
        self.local_upload = self.node.get_attribute("local_upload").get_bool()

        self.should_delete_render_package = (
            self.node.get_attribute("clean_up_render_package").get_bool()
            and self.local_upload
        )

        self.force_upload = self.node.get_attribute("force_upload").get_bool()
        self.upload_only = self.node.get_attribute("upload_only").get_bool()
        self.project = self._get_project()
        self.notifications = self._get_notifications()
        self.tokens = self._set_tokens()

        self.jobs = []
        for node in self.nodes:
            job = Job(node, self.tokens, self.render_package_path)
            self.jobs.append(job)

    def _get_project(self):
        """Get the project from the attr.

        Get its ID in case the current project is no longer in the list
        of projects at conductor, throw an error.
        """


        projects = coredata.data().get("projects")
        project_att = self.node.get_attribute("conductor_project_name")
        label = project_att.get_applied_preset_label()

        found = next((p for p in projects if p== label), None)
        if not found:
            msg = 'Cannot find project "{}" at Conductor. Please ensure the PROJECT dropdown contains a valid project.'.format(label)
            raise ConductorError(msg)

        return found

    def _get_notifications(self):
        """Get notification prefs."""
        if not self.node.get_attribute("notify").get_bool():
            return None

        emails = self.node.get_attribute("email_addresses").get_string()
        return [email.strip() for email in emails.split(",") if email.strip()]

    def _set_tokens(self):
        """Env tokens are variables to help the user build expressions.

        The user interface has fields for strings such as job title,
        task command. The user can use these tokens with <angle brackets> to build those strings. Tokens at the Submission
        level are also available in Job level fields, and likewise
        tokens at the Job level are available in Task level fields.
        """
        tokens = {}

        pdir_val = ix.application.get_factory().get_vars().get("PDIR").get_string()

        tokens["ct_pdir"] = '"{}"'.format(Path(pdir_val).posix_path(with_drive=False))

        tokens["ct_temp_dir"] = "{}".format(self.tmpdir.posix_path(with_drive=False))
        tokens["ct_submitter"] = self.node.get_name()

        tokens["ct_render_package"] = '"{}"'.format(
            self.render_package_path.posix_path(with_drive=False)
        )

        tokens["ct_project"] = self.project 

        return tokens

    def _get_render_package_path(self):
        """
        Calc the path to the render package.

        The name is not always known until
        preview/submission time because it is based on the filename. 
        What this means, practically, is that it won't
        show up in the extra uploads window along with
        other dependencies when the glob or smart-scan button is pushed.
        It will however always show up in the preview window.

        We replace spaces in the filename because of a bug in Clarisse
        https://www.isotropix.com/user/bugtracker/376

        Returns:
            string: path
        """
        current_filename = ix.application.get_current_project_filename()
        path = os.path.splitext(current_filename)[0]

        path = os.path.join(
            os.path.dirname(path), os.path.basename(path).replace(" ", "_")
        )

        try:
            return Path("{}.ct.project".format(path))
        except ValueError:
            msg= 'Cannot create a submission from this file: "{}". Has it ever been saved?'.format(current_filename)
            raise ConductorError(msg)

    def get_args(self):
        """
        Prepare the args for submission to conductor.

        Returns:
            list: list of dicts containing submission args per job.
        """

        result = []
        submission_args = {}

        submission_args["local_upload"] = self.local_upload
        submission_args["upload_only"] = self.upload_only
        submission_args["force"] = self.force_upload
        submission_args["project"] = self.project
        submission_args["notify"] = self.notifications

        for job in self.jobs:
            args = job.get_args(self.upload_only)
            args.update(submission_args)
            result.append(args)
        return result

    def submit(self):
        """
        Submit all jobs.

        Returns:
            list: list of response dictionaries, containing response codes
            and descriptions.
        """

        submission_args = self.get_args()
        self._before_submit()

        do_submission, submission_args = self.legalize_upload_paths(submission_args)
        results = []
        if do_submission:

            for job_args in submission_args:
                try:
                    remote_job = conductor_submit.Submit(job_args)
                    response, response_code = remote_job.main()
                    results.append({"code": response_code, "response": response})
                except BaseException:
                    results.append(
                        {
                            "code": "undefined",
                            "response": "".join(
                                traceback.format_exception(*sys.exc_info())
                            ),
                        }
                    )
            for result in results:
                ix.log_info(result)
        else:
            return [{"code": "undefined", "response": "Submission cancelled by user"}]

        self._after_submit()
        return results

    def _before_submit(self):
        """
        Prepare the project files that will be shipped.

        We first write out the current project file. 
        
        Then (on Windows) we find additional referenced project 
        files and adjust paths in all of them so they may be 
        rendered on linux render nodes.
        """
        self.write_render_package()


    def write_render_package(self):
        """
        Write a package suitable for rendering.

        A render package is a project file with a special name.
        """

        app = ix.application
        clarisse_window = app.get_event_window()

        self._before_write_package()
        current_filename = app.get_current_project_filename()
        current_window_title = clarisse_window.get_title()

        package_file = self.render_package_path.posix_path()
        with cu.disabled_app():
            success = ix.application.save_project(package_file)
            ix.application.set_current_project_filename(current_filename)
            clarisse_window.set_title(current_window_title)

        self._after_write_package()

        if not success:
            msg = "Failed to export render package {}".format(package_file)
            raise ConductorError(msg)
            
        ix.log_info("Wrote package to {}".format(package_file))

        if cu.is_windows():
            ix.log_info("Windows path adjustments")
            self._linuxify_project_references()
            ix.log_info("Linuxified prioject references")
            self._linuxify_render_package()
            ix.log_info("Linuxified render project file")
        else:
            ix.log_info("Not using Windows")

        return package_file

    def _linuxify_project_references(self):
        """
        Make copies of project files on Windows that are suitable for linux.

        Copies have the special conductor extension: ".ct.project".
        We convert all paths to posix. We also adjust references to
        other projects to point to the ".ct.project" version, because the 
        .ct.project version will also be linuxified.

        """
        ix.log_info("Adjust project references for Linux")
        paths = PathList()

        contexts = ix.api.OfContextSet()
        ix.application.get_factory().get_root().resolve_all_contexts(contexts)
        for context in contexts:
            if context.is_reference() and not context.is_disabled():
                try:
                    filename = context.get_attribute("filename").get_string()
                    if filename.endswith(".project"):
                        paths.add(filename)
                except ValueError as ex:
                    msg=  "{} - while resolving reference {}.filename = {}".format( str(ex), str(context), filename )
                    raise ConductorError(msg)

        # Now we have a list of all filenames that need to be adjusted.
        # So we rewrite each file, and any references to other files they may contain.
        PATH_LINE_REGEX = _get_path_line_regex()
        for path in paths:
            ix.log_info("Adjust paths in {}".format(path.posix_path()))
            self._linuxify_file(path.posix_path(), PATH_LINE_REGEX)

    def _linuxify_render_package(self):
        """
        Adjust reference pasths for windows.
        """
        PATH_LINE_REGEX = _get_path_line_regex()
        temp_path = os.path.join(
            tempfile.gettempdir(), next(tempfile._get_candidate_names())
        )
        shutil.copy2(self.render_package_path.posix_path(), temp_path)

        os.remove(self.render_package_path.posix_path())
        self._linuxify_file(
            temp_path, PATH_LINE_REGEX, self.render_package_path.posix_path()
        )

    def _linuxify_file(self, filename, path_regex, dest_path=None):
        """
        Fix paths for one file.

        If the file already has the .ct.project extension, replace it too.
        """
        out_filename = dest_path
        if not dest_path:
            out_filename = re.sub(
                PROJECT_EXTENSION_REGEX, CT_PROJECT_EXTENSION, filename
            )

        with open(out_filename, "w+") as outfile:
            with open(filename, "r+") as infile:
                for line in infile:
                    outfile.write(self._replace_path(line, path_regex))

    def _replace_path(self, line, path_regex):
        """
        Detect paths in the line of text and make a replacement.

        Args:
            line (string): line from the file.

        Returns:
            string: The line, possibly with replaced path
        """

        match = re.match(path_regex, line)
        if match:
            path = Path(match.group(1), no_expand=True).posix_path(with_drive=False)
            path = re.sub(PROJECT_EXTENSION_REGEX, CT_PROJECT_EXTENSION, path)
            return line.replace(match.group(1), path)

        return line

    def legalize_upload_paths(self, submission_args):
        """
        Alert the user of missing files. If the user doesn't want to continue
        with missing files, the result will be False. Otherwise it will be True
        and the potentially adjusted args are returned.

        Args:
            submission_args (list): list of job args.

        Returns:
           tuple (bool, adjusted args):
        """
        missing_files = []

        for job_args in submission_args:
            existing_files = []
            for path in job_args["upload_paths"]:
                if os.path.exists(path):
                    existing_files.append(path)
                else:
                    missing_files.append(path)

            job_args["upload_paths"] = existing_files
        missing_files = sorted(list(set(missing_files)))
        if not missing_files_ui.proceed(missing_files):
            return (False, [])
        return (True, submission_args)

    def _before_write_package(self):
        """
        Prepare to write render package.
        """
        if self.localize_before_ship:
            _localize_contexts()
            _remove_conductor()

        self._prepare_temp_directory()
        self._copy_system_dependencies_to_temp()

    def _prepare_temp_directory(self):
        """
        Make sure the temp directory has a conductor subdirectory.
        """
        tmpdir = self.tmpdir.posix_path()
        try:
            os.makedirs(tmpdir)
        except OSError as ex:
            if not (ex.errno == errno.EEXIST and os.path.isdir(tmpdir)):
                raise

    def _copy_system_dependencies_to_temp(self):
        """
        Copy over all system dependencies to a tmp folder.

        Wrapper scripts, config files etc. The clarisse.cfg file is special. See
        ../clarisse_config.py
        """
        for entry in deps.system_dependencies():
            if os.path.isfile(entry["src"]):
                if entry["src"].endswith(".cfg"):
                    safe_config = ccfg.legalize(entry["src"])
                    with open(entry["dest"], "w") as dest:
                        dest.write(safe_config)
                    ix.log_info(
                        "Copy with mods {} to {}".format(entry["src"], entry["dest"])
                    )
                else:
                    ix.log_info("Copy {} to {}".format(entry["src"], entry["dest"]))
                    shutil.copy(entry["src"], entry["dest"])

    def _after_submit(self):
        """Clean up, and potentially other post submission actions."""
        self._delete_render_package()

    def _after_write_package(self):
        """
        Runs operations after saving the render package.

        If we did something destructive, like localize contexts, then
        a backup will have been saved and we now reload it. This strategy
        is used because Clarisse's undo is broken when it comes to
        undoing context localization.
        """
        if self.localize_before_ship:
            self._revert_to_saved_scene()

    def _delete_render_package(self):
        """
        Delete the render package from disk if the user wants to.
        """
        if self.should_delete_render_package:
            render_package_file = self.render_package_path.posix_path()
            if os.path.exists(render_package_file):
                os.remove(render_package_file)

    def _revert_to_saved_scene(self):
        """
        If contexts were localized, we will load the saved scene.
        """
        with cu.waiting_cursor():
            with cu.disabled_app():
                ix.application.load_project(self.project_filename)

    @property
    def node_name(self):
        """node_name."""
        return self.node.get_name()

    @property
    def filename(self):
        """filename."""
        return ix.application.get_current_project_filename()

    def has_notifications(self):
        """has_notifications."""
        return bool(self.notifications)

    @property
    def email_addresses(self):
        """email_addresses."""
        if not self.has_notifications():
            return []
        return self.notifications["email"]["addresses"]



"""Handle button presses to submit and preview jobs.

Preview, open a window containing the submission script JSON, and
eventually also the structure of the submission and the JSON objects
that will be sent to Conductor.

Submit, send jobs straight to Conductor.
"""
import os
import ix
from cioclarisse import utils
from cioclarisse.utils import ConductorError
from cioclarisse.scripted_class import preview_ui
from cioclarisse.scripted_class.submission import Submission
from ciocore.gpath_list import PathList
from ciocore import data as coredata
from cioclarisse import const as k
SUCCESS_CODES_SUBMIT = [201, 204]

SUBMIT_DIRECT = 0
PREVIEW_FIRST = 1
WRITE_PACKAGE_ONLY = 2

SAVE_STATE_UNMODIFIED = 0
SAVE_STATE_CANCELLED = 1
SAVE_STATE_SAVED = 2
SAVE_STATE_NO = 3
SAVE_STATE_DONT_CARE = 4


def check_need_save(node, which):
    """
    Check if the current project is modified and proposed to save it.

    If user chooses to save, then save it and return the filename along
    with an enum response.


    Args:
        node (ConductorJob):
        which (Enum): Why we are checking.
            SUBMIT_DIRECT, PREVIEW_FIRST, WRITE_PACKAGE_ONLY

    Returns:
        tuple(Enum, Filename):
            SAVE_STATE_UNMODIFIED
            SAVE_STATE_CANCELLED
            SAVE_STATE_SAVED
            SAVE_STATE_NO
            SAVE_STATE_DONT_CARE

    """

    if not node.get_attribute("localize_contexts").get_bool():
        return SAVE_STATE_DONT_CARE, None
    if which == SUBMIT_DIRECT:
        msg = "Save the project?\nYou must save the project if you wish to submit."
    elif which == PREVIEW_FIRST:
        msg = "Save the project?\nClick YES to preview with the option to submit.\nClick NO to preview only."
    else:  # WRITE_PACKAGE_ONLY
        msg = "Save the project?\nYou must save the project if you wish to export a render package."

    response = ix.api.AppDialog.cancel()
    app = ix.application
    if not ix.application.is_project_modified():
        return SAVE_STATE_UNMODIFIED, None
    cwin = ix.application.get_event_window()
    box = ix.api.GuiMessageBox(
        app, 0, 0, "Conductor Information - project not saved!", msg
    )
    x = (2 * cwin.get_x() + cwin.get_width() - box.get_width()) / 2
    y = (2 * cwin.get_y() + cwin.get_height() - box.get_height()) / 2
    box.resize(x, y, box.get_width(), box.get_height())
    if which == SUBMIT_DIRECT or which == WRITE_PACKAGE_ONLY:
        box.set_style(ix.api.AppDialog.STYLE_YES_NO)
    else:
        box.set_style(ix.api.AppDialog.STYLE_YES_NO_CANCEL)

    box.show()
    response = box.get_value()
    box.destroy()

    if response.is_cancelled():
        return SAVE_STATE_CANCELLED, None
    if response.is_no():
        return SAVE_STATE_NO, None

    # response is yes
    current_filename = ix.application.get_current_project_filename()

    if current_filename == "":
        current_filename = "untitled"
    filename = ix.api.GuiWidget.save_file(
        app, current_filename, "Save Scene File...", "Project Files\t*." + "project"
    )
    if filename:
        ix.application.save_project(filename)
        return SAVE_STATE_SAVED, filename
    else:
        return SAVE_STATE_CANCELLED, None


def submit(*args):
    """
    Validate and submit directly.
    """
    node = args[0]

    state, fn = check_need_save(node, SUBMIT_DIRECT)
    if state not in [SAVE_STATE_UNMODIFIED, SAVE_STATE_SAVED, SAVE_STATE_DONT_CARE]:
        ix.log_warning("Submission cancelled.")
        return
    _validate(node)

    with utils.waiting_cursor():
        submission = Submission(node)
        responses = submission.submit()

    preview_ui.show_submission_responses(responses)


def preview(*args):
    """
    Validate and show the script in a panel.

    Submission can be invoked from the preview panel.
    """
    node = args[0]

    state, _ = check_need_save(node, PREVIEW_FIRST)
    if state == SAVE_STATE_CANCELLED:
        ix.log_warning("Preflight cancelled.")
        return
    can_submit = state in [
        SAVE_STATE_UNMODIFIED,
        SAVE_STATE_SAVED,
        SAVE_STATE_DONT_CARE,
    ]

    _validate(node)
    with utils.waiting_cursor():
        submission = Submission(node)
    preview_ui.build(submission, can_submit=can_submit)


def _validate(node):
    print "_validate_images"
    _validate_images(node)
    print "_validate_packages"
    _validate_packages(node)
    print "_validate_project"
    _validate_project(node)


def _validate_images(node):
    """
    Check that images or layers are present and set up to be rendered.

    Also check that the when there are many output paths, their common path is
    not the filesystem root, as this will be the submission's output_path
    property.

    Args:
        node (ConductorJob): Node
    """
    images = ix.api.OfObjectArray()
    node.get_attribute("images_and_layers").get_values(images)
    out_paths = PathList()


    num_images = images.get_count()
    print num_images    
    if not num_images:
        raise(BaseException("No render outputs added. Please add images and/or layers in the Submitter"))

    for image in images:
        if not image.get_attribute("render_to_disk").get_bool():
            msg = "Image does not have render_to_disk attribute set: {}".format(image.get_full_name())
            raise ConductorError(msg)

        save_path = image.get_attribute("save_as").get_string()
        if not save_path:
            msg = "Image save_as path is not set: {}".format(image.get_full_name())
            raise ConductorError(msg)

        if save_path.endswith("/"):
            msg =  "Image save_as path must be a filename,   not a directory: {}".format( image.get_full_name() )
            raise ConductorError(msg)
 

        directory = os.path.dirname(save_path)
        out_paths.add(directory)

    if not out_paths:
        msg = "There are no images or layers with Save to Disk turned on."
        raise ConductorError(msg)

    common_path = out_paths.common_path()

    paths = "\n".join(p.posix_path() for p in out_paths)

    if common_path.depth == 0:
        msg = "Your output files should be rendered to a common subfolder.  Not the filesystem root. {}\n{}".format(     common_path.posix_path(), paths )
        raise ConductorError(msg)

def _validate_packages(obj):
    """
    Make sure a Clarisse version is selected.

    Args:
        obj (ConductorJob):
    """
    if not (
        obj.get_attribute("clarisse_version")
        .get_applied_preset_label()
        .startswith("clarisse")
    ):
        raise ConductorError("No Clarisse package selected.")
 

def _validate_project(obj):
    """
    Check the project is set.

    Args:
        obj (ConductorJob):
    """
    projects = coredata.data().get("projects")

    project_att = obj.get_attribute("conductor_project_name")
    label = project_att.get_applied_preset_label()
    if not label or label == k.NOT_CONNECTED :
        msg = 'Project is not set for "{}".'.format(obj.get_name())
        raise ConductorError(msg)
    try:
        next(p for p in projects if str(p) == label)
    except StopIteration:
        msg = 'Cannot find project "{}" at Conductor. Please ensure the PROJECT dropdown contains a valid project.'.format(label)
        raise ConductorError(msg)
 


"""
Collect dependencies
"""

import os
import re

import cioclarisse.utils as cu
import ix
from cioclarisse.scripted_class import frames_ui
from ciocore.gpath import Path
from ciocore.gpath_list import PathList
from ciocore.sequence import Sequence
from cioclarisse.utils import ConductorError

RX_HASH = re.compile(r"#+")
RX_FRAME = re.compile(r"\$(\d?)F")
RX_FRAME_FORMAT = re.compile(r"\{frame:\d*d}")
RX_UDIM = re.compile(r"<UDIM>")
GLOB = 1
SMART = 2



# ct_node: wrapper around cnode. Allows us to add flags and so on.

# cioprep.py: pre render script that operates in the Clarisse. Among other
# things, it emoves drive letters on win.
CONDUCTOR_SCRIPTS = ["cioprep.py", "ciocnode"]

# clarisse.cfg: A copy of the users config file.
CLARISSE_CFG_FILENAME = "clarisse.cfg"


def system_dependencies():
    """
    Provides a list of system files to be sent to the render node.

    These will be copied to a directory in preparation for uploading.

    This is part of a strategy to satisfy 2 constraints.
    1. Dont store special logic on the sidecar.
    2. Don't make the render command un-runnable on the local machine.

    See docs in ciocnode and cioprep for more info.

    Returns:
        list: Each element is a source/destination pair of paths.
        [
            {"src": "/some/path.ext", "dest": "/other/path.ext"},
            ...
        ]
    """

    result = []
    conductor_scripts_directory = os.path.join(os.path.dirname(os.path.dirname(__file__)), "scripts")

    conductor_tmp_dir = os.path.join(
        ix.application.get_factory().get_vars().get("CTEMP").get_string(), "conductor"
    )

    for script in CONDUCTOR_SCRIPTS:
        src_path = Path(os.path.join(conductor_scripts_directory, script)).posix_path()
        dest_path = Path(os.path.join(conductor_tmp_dir, script)).posix_path()

        result.append({"src": src_path, "dest": dest_path})

    config_dir = (
        ix.application.get_factory()
        .get_vars()
        .get("CLARISSE_USER_CONFIG_DIR")
        .get_string()
    )

    config_src_file = Path(os.path.join(config_dir, CLARISSE_CFG_FILENAME)).posix_path()
    config_dest_file = Path(
        os.path.join(conductor_tmp_dir, CLARISSE_CFG_FILENAME)
    ).posix_path()

    result.append({"src": config_src_file, "dest": config_dest_file})

    return result


def _get_system_dependencies():
    """
    Extracts the destination side of system dependency files.

    Returns:
        PathList: list of system files to be uploaded
    """
    result = PathList()
    for entry in system_dependencies():
        try:
            result.add(entry["dest"])
        except ValueError as ex:
            msg = "{} - while resolving system_dependency: {}".format( str(ex), entry["dest"])
            raise ConductorError(msg)

    return result


def collect(obj):
    """
    Collect ALL upload files in preparation for submission.

    Args:
        obj (ConductorJob): The item whose attributes define the scan.
    Returns:
        PathList: All file dependencies.
    """
    result = PathList()

    # policy: NONE, GLOB, SMART_SEQUENCE
    policy = obj.get_attribute("dependency_scan_policy").get_long()

    # If we are not localizing reference contexts, then we need to add the
    # referenced project files to the dependencies list.
    include_references = not obj.get_attribute("localize_contexts").get_bool()

    result.add(*_get_system_dependencies())
    result.add(*_get_extra_uploads(obj))

    result.add(*get_scan(obj, policy, include_references))

    # tell the list to look on disk and expand  any "*" or <UDIM> in place
    result.glob()

    return result


def _get_extra_uploads(obj):
    """
    Collects any files specified through the extra uploads window.

    They are stored in a list attribute on the ConductorJob item.

    Args:
        obj (ConductorJob): item being processed.

    Returns:
        PathList: Collected paths.
    """
    result = PathList()
    extras_attr = obj.get_attribute("extra_uploads")
    paths = ix.api.CoreStringArray()
    extras_attr.get_values(paths)
    for path in paths:
        try:
            result.add(path)
        except ValueError as ex:
            msg = "{} - while resolving extra upload path: {}".format(str(ex), path)
            raise ConductorError(msg)

    return result


def get_scan(obj, policy, include_references=True):
    """
    Scan all path attrs for dependencies according to the given policy.

    If policy is not None: First replace all UDIM tags with a "*" so they may be
    globbed.

    File sequences may be resolved one of 2 ways:
    1. If policy is GLOB, then hashes in filenames will be replaced by a "*" and
        globbed later.
    2. If policy is SMART then for each filename, we look at its sequence
        definition and calculate the frames that will be needed for the frame
        range being rendered.

    Args:
        obj (ConductorJob): Item being processed.
        policy (Enum): NONE, GLOB, SMART
        include_references (bool, optional): Whether to scan for references. Defaults to True.

    Returns:
        PathList: Collected paths
    """
    result = PathList()

    if not policy:
        return result

    for attr in ix.api.OfAttr.get_path_attrs():
        if _should_ignore(attr):
            continue

        if attr.is_expression_enabled() and attr.is_expression_activated():
            filename = _evaluate_static_expression(attr)
        else:
            filename = attr.get_string()

        if not filename:
            continue

        # always glob for udims.
        filename = RX_UDIM.sub("*", filename)

        if policy == GLOB:
            # replace all frame identifiers with "*" for globbing later
            try:
                result.add(re.sub((r"(#+|\{frame:\d*d})"), "*", filename))
            except ValueError as ex:
                msg =  "{} - while resolving path: {} = {}".format(     str(ex), str(attr), filename )
                raise ConductorError(msg)

        else:  # SMART
            filenames = _smart_expand(obj, attr, filename)
            try:
                result.add(*filenames)
            except ValueError as ex:
                msg = "{} - while resolving path: {} = {}".format(str(ex), str(attr), filename)
                raise ConductorError(msg)
                 
    # We ignored references in _should_ignore for 2 reasons.
    # 1. We only want them if we are not localizing them
    #    (include_references==True).
    # 2. Getting ref contexts from OfAttr.get_path_attrs() is buggy so it's best
    #    to get them through the root context with resolve_all_contexts()

    # On Windows, we make a new extension for all project reference paths (".ct.project")
    # because we will be replacing them with a linuxified version of the file.
    if include_references:
        refs = _scan_for_references()
        if cu.is_windows():
            for ref in refs:
                if ref.endswith(".project"):
                    result.add(
                        re.sub(
                            r"(\.ct\.project|\.project)",
                            ".ct.project",
                            ref.posix_path(),
                        )
                    )
                else:
                    result.add(ref)
        else:
            result.add(*refs)

    return result


def _scan_for_references():
    result = PathList()
    contexts = ix.api.OfContextSet()
    ix.application.get_factory().get_root().resolve_all_contexts(contexts)
    for context in contexts:
        if context.is_reference() and not context.is_disabled():
            try:
                filename = context.get_attribute("filename").get_string()
                result.add(filename)
            except ValueError as ex:
                msg =  "{} - while resolving reference {}.filename = {}".format(str(ex), str(context), filename)
                raise ConductorError(msg)

    return result


def _is_project_reference(attr):
    """
    Does the attribute reference a clarisse project?

    If so, we don't include it in the depedency scan because ultimately it will
    be made local. Check the extension, because Alembic and pixar formats are
    also possible.

    Args:
        attr (OfAttr): Attribute to query

    Returns:
         bool: Is the attribute a reference
    """
    obj = attr.get_parent_object()
    if obj.get_class_name() == "Generic":
        if obj.get_name() == "__context_options__":
            path = attr.get_string()
            _, ext = os.path.splitext(path)
            if ext == ".project":
                return True
    return False


def _should_ignore(attr):
    """
    Is the attribute needed

    Args:
        attr (OfAttr): Attribute to query

    Returns:
        bool: True if the path at this attribute should be ignored
    """
    if attr.get_parent_object().is_disabled():
        return True
    if attr.is_private():
        return True
    hint = ix.api.OfAttr.get_visual_hint_name(attr.get_visual_hint())
    if hint == "VISUAL_HINT_FILENAME_SAVE":
        return True
    if _is_project_reference(attr):
        return True
    return False


def _evaluate_static_expression(target_attr):
    """
    Evaluate the static part of an attribute containing an expression.

    Assume expression is in active state, (otherwise we wouldn't be here).

    We are doing this because we want literal paths that are needed for the
    upload. The easiest and most reliable way to evaluate variables in the
    expression would be to use get_string(). However, if we do that, it will
    evaluate time varying variables like $3F for the current frame only.

    So to keep the time varying varables variable, we replace them with python named format strings
    like {frame:0nd} where n is the padding.

    NOTE: It does not resolve hash placeholders like ### so we can leave them as is.
    NOTE 2: Can't remember why I used {frame:04d} rather than change to ####.

    $CDIR/bugs_seq/bugs.####.jpg becomes
    /Users/julian/projects/fish/clarisse/bugs_seq/bugs.####.jpg

    $CDIR/bugs_seq/bugs.$4F.jpg becomes
    /Users/julian/projects/fish/clarisse/bugs_seq/bugs.{frame:04d}.jpg

    With these time varying placeholders left intact, we can expand correctly
    later using either the SMART scan policy, or a simple GLOB.

    Args:
        target_attr (OfAttr): Attribute to query,

    Returns:
        string: The attribute value with all $variables resolved except those
        that rely on time.
    """
    orig_expr = target_attr.get_expression()

    static_expr = re.sub(
        r"\$(\d?)F",
        lambda match: "{{frame:0{:d}d}}".format(int(match.group(1) or 0)),
        orig_expr,
    )

    target_attr.set_expression(static_expr)
    result = target_attr.get_string()
    target_attr.set_expression(orig_expr)
    return result


def _smart_expand(obj, attr, filename):
    """
    Expand filenames when policy is SMART.

    Takes into account the frame range to be rendered, and the frame
    specification of any source animated maps along with their offset. Then
    resolves filenames for that intersection of sequences.

    Args:
        obj (ConductorJob):  Item being processed. attr (OfAttr): Attribute to
        query.
        filename (string): File template to expand. It may have hashes
        and/or frame_format {frame:02d} style placeholders to represent frame
        ranges

    Returns:
        list(strings): Expanded filenames
    """

    result = []
    main_seq = frames_ui.main_frame_sequence(obj)

    sequences = None

    do_resolve_hashes = RX_HASH.search(filename)
    do_resolve_frame_format = RX_FRAME_FORMAT.search(filename)
    # If a filename had both of the above (pretty rare) then
    # we need to expand the hashes first, but also have a record
    # of what frame numbers (in the timeline) relate to what files,
    # because remember that hashes can represent a sequence with an
    # offset. So after the hash expansion is done, go and resolve
    # the corresponding {frame:02d} style placeholders.
    # To this end, the function _attribute_sequence() returns both
    # the Sequence representing the files, and the Sequence
    # representing the render.
    # sequences["attr_sequence"]
    # sequences["render_sequence"]
    if do_resolve_hashes:
        sequences = _attribute_sequence(attr, main_seq)
        if sequences:
            result = sequences["attr_sequence"].expand(filename)
            if do_resolve_frame_format:
                result = sequences["render_sequence"].expand_format(*result)
    elif do_resolve_frame_format:
        # do_resolve_frame_format ONLY for filename
        result = main_seq.expand_format(filename)
    else:
        # There are no hashes (that intersect) or frame_format_expr
        # components in the filename. So just return the filename
        result = [filename]
    return result


def _attribute_sequence(attr, intersector):
    """
    Get the sequence associated with a filename attribute.

    Many attributes have an associated sequence_mode attribute, which when set
    to 1 signifies varying frames and makes available start, end, and offset
    attributes to help specify the sequence.

    We work out the intersection of that sequence with the main sequence because
    during dependency scanning, we can optimize the number of frames to upload
    if we use only those frames specified in the sequence attribute, and
    intersect them with the frame range specified in the job node.

    Args:
        attr (OfAttr): Attribute to query,
        intersector (Sequence): The main sequence as defined by the ConductorJob
        frames override, or by the images being rendered.

    Returns:
        dict:
            The sequence defined by the attribute intersected with the main sequence.
            The sequence the sequence while at the render frames.
    """

    obj = attr.get_parent_object()
    mode_attr = obj.attribute_exists("sequence_mode")
    if not (mode_attr and mode_attr.get_long()):
        msg = "Attribute is not in sequence mode"
        raise ConductorError(msg)
   
    global_frame_rate = ix.application.get_prefs(
        ix.api.AppPreferences.MODE_APPLICATION
    ).get_long_value("animation", "frames_per_second")
    attr_frame_rate = obj.get_attribute("frame_rate").get_long()
    if not attr_frame_rate == global_frame_rate:
        msg =  "Can't get attribute sequence when global fps is different from fps on the attribute."
        raise ConductorError(msg)

    start = obj.get_attribute("start_frame").get_long()
    end = obj.get_attribute("end_frame").get_long()

    # If there's a frame offset on the attribute, then we need to
    # do the intersection in the context of that offset.
    offset = obj.get_attribute("frame_offset").get_long()

    seq = Sequence.create(start, end, 1).offset(offset)
    seq = seq.intersection(intersector)


    if not seq:
        # The attribute doesn't intersect the render frames
        return
    # Make a copy of the sequence while at the render frames
    render_seq = Sequence.create(str(seq))
    attr_seq = seq.offset(-offset)

    return {"attr_sequence": attr_seq, "render_sequence": render_seq}



# from __future__ import unicode_literals

# import argparse
import platform
import os
import re
import sys
import errno
from shutil import copy2
 


CIO_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
CONFIG_FILENAME = "clarisse.env"
PLATFORM = sys.platform

def main():
    sys.stdout.write("Running Clarisse post install script\n")

    if not PLATFORM in ["darwin", "win32", "linux"]:
        sys.stderr.write("Unsupported platform: {}".format(PLATFORM))
        sys.exit(1)
    transform_env_files()
    sys.stdout.write("Completed Clarisse post install script\n")

    sys.exit(0)


def transform_env_files():
    """Find env files below the Clarisse dir and add a Conductor location variable."""
    if PLATFORM == "darwin":
        root = os.path.expanduser(os.path.join("~","Library","Preferences","Isotropix","Clarisse"))
    elif PLATFORM == "linux":
        root = os.path.expanduser(os.path.join("~",".isotropix","clarisse"))
    else:  # windows
        appdata=os.environ.get("APPDATA")
        root = os.path.join(appdata, "Isotropix", "Clarisse")
    transform_tree(root)


def transform_tree(root):
    """Walk the filesystem and transform env files."""
    root, dirs, filenames = next(os.walk(root), (None,[], []))
    for fn in filenames:
        if fn == CONFIG_FILENAME:
            transform_env_file(os.path.join(root, fn))
    for dr in dirs:
        transform_tree(os.path.join(root, dr))

def transform_env_file(env_file):
    lines = []
    with open(env_file) as fn:
        for line in fn:
            if line.startswith("CIO_DIR"):
                lines.append("CIO_DIR={}".format(CIO_DIR))
            else:
                lines.append(line.rstrip())

    with open(env_file, "w") as fn:
        for line in lines:
            fn.write(u"{}\n".format(line))
    sys.stdout.write("Wrote Conductor additions to : {}\n".format(env_file))

if __name__ == '__main__':
    main()
